/*************************************************************************
 *
 *  C-to-C Translator
 *
 *  Adapted from Clean ANSI C Parser
 *  Eric A. Brewer, Michael D. Noakes
 *
 *************************************************************************/
/*
 * Copyright (c) 2000 Massachusetts Institute of Technology
 * Copyright (c) 2000 Matteo Frigo
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "ast.h"

/* Set to 1 to enable parser debugging.  Also requires YACC flags */
extern int yydebug;

GLOBAL const char *Executable;
GLOBAL const float VersionNumber = 0.6;
GLOBAL const char *const VersionDate = __DATE__;
GLOBAL const char *PhaseName = "???";
GLOBAL const char *InputFileRoot;

#define MAX_LINK_FILES  200

/* values 1-5 are valid; 5 == all */
GLOBAL int WarningLevel = WARNING_LEVEL;

GLOBAL List *Program;

GLOBAL extern FILE *yyin;	/* file pointer used by the lexer */
PRIVATE const char *input_file = NULL;
PRIVATE const char *output_file = NULL;
PRIVATE const char *stdin_name = NULL;

/* global flags */
GLOBAL Bool DebugLex = FALSE;
GLOBAL Bool PrintSymTables = FALSE;
GLOBAL Bool TrackInsertSymbol = FALSE;
GLOBAL Bool PrintAST = FALSE;
GLOBAL Bool PrintLineOffset = FALSE;
GLOBAL Bool IgnoreLineDirectives = FALSE;
GLOBAL Bool ANSIOnly = FALSE;
GLOBAL Bool PrintLiveVars = FALSE;

GLOBAL Bool Parse = TRUE;
GLOBAL Bool SemanticCheck = TRUE;
GLOBAL Bool Analyze = TRUE;
GLOBAL Bool DoHint = FALSE;
GLOBAL Bool Transform = TRUE;
GLOBAL Bool GenerateOutput = TRUE;
GLOBAL Bool FormatReadably = FALSE;
GLOBAL Bool GenerateSegments = FALSE;

PRIVATE void print_version_info(FILE *out, const char *pre)
{
     fprintf(out, "%sVersion %.02f (%s)\n",
	     pre, VersionNumber, VersionDate);
     exit(0);
}

PRIVATE void print_copyright(FILE *out, const char *pre)
{
     static const char *lines[] = {
	  "Copyright (c) 1994-2000 MIT Laboratory for Computer Science\n",
	  "Copyright (c) 2000 Matteo Frigo\n",
	  "Portions Copyright (c) 1989, 1990 James A. Roskind\n",
	  NULL
     };
     int i;

     for (i = 0; lines[i] != NULL; i++) {
	  fputs(pre, out);
	  fputs(lines[i], out);
     }
     exit(0);
}

PRIVATE void usage(Bool print_all_options, int exitcode)
{
     fprintf(stderr, "Usage: %s [options] [file]\n", Executable);

     fprintf(stderr,
	     "\n"
	     "Translates Cilk 5.3 source in file%s to C output in file%s.\n"
	     "If file%s is omitted, uses standard input and standard output.\n"
	     "\n",
	     INPUT_SUFFIX, OUTPUT_SUFFIX, INPUT_SUFFIX);

     fprintf(stderr, "General Options:\n");
     fprintf(stderr,
	     "\t-help              Print this description\n");
     fprintf(stderr,
	     "\t-options           Print all options\n");
     fprintf(stderr,
	     "\t-copy              Print the copyright information\n");
     fprintf(stderr,
	     "\t-v                 Print version information\n");

     if (print_all_options) {
	  fprintf(stderr, "Phase Options:\n");
	  fprintf(stderr,
		  "\t-nosem             Don't semantic-check\n");
	  fprintf(stderr,
	       "\t-noanalyze         Don't perform dataflow analysis\n");
	  fprintf(stderr,
		  "\t-notrans           Don't transform syntax tree\n");
	  fprintf(stderr,
		  "\t-noprint           Don't write C output\n");

	  fprintf(stderr, "Output Options:\n");
	  fprintf(stderr,
		  "\t-o <name>          Write C output to <name>\n");
	  fprintf(stderr,
		  "\t-N                 Don't emit line directives\n");

	  fprintf(stderr, "Warning Options:\n");
	  fprintf(stderr,
		  "\t-ansi              Disable GCC extensions and undefine __GNUC__\n");
	  fprintf(stderr,
		  "\t-W<n>              Set warning level; <n> in 1-5. Default=%d\n",
		  WARNING_LEVEL);
	  fprintf(stderr,
		  "\t-Wall              Same as -W5\n");
	  fprintf(stderr,
		  "\t-il                Ignore line directives (use actual line numbers)\n");
	  fprintf(stderr,
		  "\t-offset            Print offset within the line in warnings/errors\n");
	  fprintf(stderr,
		  "\t-name <x>          Use stdin with <x> as filename in messages\n");

	  fprintf(stderr, "Debugging Options:\n");
	  fprintf(stderr,
		  "\t-lex               Show lexical tokens\n");
	  fprintf(stderr,
		"\t-yydebug           Track parser stack and actions\n");
	  fprintf(stderr,
		  "\t-insert            Track symbol creation\n");
	  fprintf(stderr,
	   "\t-sym               Print out symbol tables after parse\n");
	  fprintf(stderr,
		  "\t-ast               Print out syntax tree (after last phase)\n");
	  fprintf(stderr,
		  "\t-live              Print live variables as cmts in C output\n");
	  fprintf(stderr, "CC Compatibility Options:\n");
	  fprintf(stderr,
	     "\t--                 Toggles ignoring of unknown options\n"
	  "\t                   (for makefile compatibility with cc)\n");

	  fprintf(stderr, "\n");
     }
     exit(exitcode);
}

PRIVATE void unknown_option(char *option)
{
     fprintf(stderr, "Unknown option: `%s'\n\n", option);
     usage(FALSE, 1);
}

/*
 * Generate a filename with a new suffix.
 * If <filename> ends with <old_suffix>, replace the suffix with <new_suffix>;
 * otherwise just append <new_suffix>. 
 */
PRIVATE const char *with_suffix(const char *filename,
				const char *old_suffix,
				const char *new_suffix)
{
     unsigned int root_len, old_len, len;
     char *newfilename;

     /* Look for old_suffix at end of filename */
     root_len = strlen(filename);
     old_len = strlen(old_suffix);
     if (root_len >= old_len &&
	 !strcmp(filename + root_len - old_len, old_suffix))
	  root_len -= old_len;

     /* Compute the length of the create filename */
     len = root_len + strlen(new_suffix) + 1;

     /* allocate the create name */
     if ((newfilename = HeapNewArray(char, len)) == NULL) {
	  printf("INTERNAL ERROR: Unable to allocate %d bytes for a filename\n",
		 len);
	  exit(-1);
     }
     strncpy(newfilename, filename, root_len);
     strcat(newfilename, new_suffix);

     return newfilename;
}

char *filename_root(const char *filename)
{
     const char *start;
     int len;
     const char *p;
     char *root;

     /* initial approximation: root == entire filename */
     start = filename;
     len = strlen(filename);

     /* remove directory prefixes one by one */
     while ((p = strchr(start, '/')) != NULL)
	  start = p + 1;

     /* remove all filename suffixes */
     if ((p = strchr(start, '.')) != NULL)
	  len = p - start;

     /* copy root out of filename */
     if ((root = HeapNewArray(char, len + 1)) == NULL) {
	  printf("INTERNAL ERROR: Unable to allocate %d bytes for a filename\n",
		 len);
	  exit(-1);
     }
     strncpy(root, start, len);
     root[len] = '\0';

     return root;
}

/***********************************************************************\
 * Handle command-line arguments
\***********************************************************************/

PRIVATE void handle_options(int argc, char *argv[])
{
     int i;

     for (i = 1; i < argc; i++) {
	  if (argv[i][0] == '-') {
	       switch (argv[i][1]) {
		   case 'a':
			if (strcmp(argv[i], "-ansi") == 0) {
			     ANSIOnly = TRUE;
			} else if (strcmp(argv[i], "-ast") == 0)
			     PrintAST = TRUE;
			else
			     unknown_option(argv[i]);
			break;

		   case 'N':
			FormatReadably = TRUE;
			break;
		   case 'W':
			if (strcmp(argv[i], "-Wall") == 0) {
			     WarningLevel = 5;
			} else {
			     int c = atoi(&argv[i][2]);
			     if (c < 0 || c > 5) {
				  unknown_option(argv[i]);
			     } else {
				  WarningLevel = c;
			     }
			}
			break;
		   case 'c':
			if (strcmp(argv[i], "-copy") == 0)
			     print_copyright(stderr, "");
			else
			     unknown_option(argv[i]);
			break;
 	           case 'h':
		        if (strcmp(argv[i], "-hint") == 0)
			     DoHint = TRUE;
			else if (strcmp(argv[i], "-h") == 0)
			     usage(FALSE, 0);
			else
			     unknown_option(argv[i]);
			break;
		   case 'i':
			if (strcmp(argv[i], "-insert") == 0) {
			     TrackInsertSymbol = TRUE;
			} else if (strcmp(argv[i], "-il") == 0) {
			     IgnoreLineDirectives = TRUE;
			} else
			     unknown_option(argv[i]);
			break;
		   case 'l':
			if (strcmp(argv[i], "-lex") == 0)
			     DebugLex = TRUE;
			else if (strcmp(argv[i], "-live") == 0)
			     PrintLiveVars = TRUE;
			else
			     unknown_option(argv[i]);
			break;
		   case 'n':
			if (strcmp(argv[i], "-name") == 0) {
			     i++;
			     if (input_file != NULL) {
				  fprintf(stderr,
					  "Multiple input files defined, using `%s'\n",
					  input_file);
			     } else {
				  stdin_name = argv[i];
			     }
			} else if (strcmp(argv[i], "-nosem") == 0) {
			     SemanticCheck = FALSE;
			} else if (strcmp(argv[i], "-notrans") == 0) {
			     Transform = FALSE;
			} else if (strcmp(argv[i], "-noprint") == 0) {
			     GenerateOutput = FALSE;
			} else if (strcmp(argv[i], "-noanalyze") == 0) {
			     Analyze = FALSE;
			} else
			     unknown_option(argv[i]);
			break;
		   case 'o':
			if (strcmp(argv[i], "-o") == 0) {
			     i++;
			     output_file = argv[i];
			} else if (strcmp(argv[i], "-offset") == 0)
			     PrintLineOffset = TRUE;
			else if (strcmp(argv[i], "-options") == 0) {
			     usage(TRUE, 0);
			     exit(0);
			} else
			     unknown_option(argv[i]);
			break;
		   case 's':
			if (strcmp(argv[i], "-sym") == 0)
			     PrintSymTables = TRUE;
			else
			     unknown_option(argv[i]);
			break;
		   case 'v':
			if (strcmp(argv[i], "-v") == 0)
			     print_version_info(stderr, "");
			else
			     unknown_option(argv[i]);
			break;
		   case 'y':
			if (strcmp(argv[i], "-yydebug") == 0)
			     yydebug = 1;
			else
			     unknown_option(argv[i]);
			break;
		   default:
			unknown_option(argv[i]);
	       }
	  } else {
	       if (input_file != NULL) {
		    fprintf(stderr, "Multiple input files defined, using `%s'\n",
			    argv[i]);
	       }
	       input_file = argv[i];
	  }
     }

     if (GenerateOutput == TRUE && input_file != NULL && output_file == NULL)
	  output_file = with_suffix(input_file, INPUT_SUFFIX, OUTPUT_SUFFIX);

}

/***********************************************************************\
 * ANSI C symbol tables
\***********************************************************************/

GLOBAL SymbolTable *Identifiers, *Labels, *Tags, *Externals;

PRIVATE void shadow_var(Node *create, Node *shadowed)
{
     /* the two are equal only for redundant function/extern declarations */
     if (create != shadowed && WarningLevel == 5) {
	  WarningCoord(5, create->coord,
		  "`%s' shadows previous declaration", VAR_NAME(create));
	  fprintf(stderr, "\tPrevious declaration: ");
	  PRINT_COORD(stderr, shadowed->coord);
	  fputc('\n', stderr);
     }
}

PRIVATE void init_symbol_tables(Bool shadow_warnings)
{
     ShadowProc shadow_proc;

     if (shadow_warnings)
	  shadow_proc = (ShadowProc) shadow_var;
     else
	  shadow_proc = NULL;

     Identifiers = NewSymbolTable("Identifiers", Nested,
				shadow_proc, (ExitscopeProc) OutOfScope);
     Labels = NewSymbolTable("Labels", Flat,
			     NULL, (ExitscopeProc) EndOfLabelScope);
     Tags = NewSymbolTable("Tags", Nested,
	    shadow_warnings ? (ShadowProc) ShadowTag : (ShadowProc) NULL,
			   NULL);
     Externals = NewSymbolTable("Externals", Flat,
				NULL, (ExitscopeProc) OutOfScope);
}

/***********************************************************************\
 * Determine input file
\***********************************************************************/

PRIVATE FILE *get_input_file(void)
{
     const char *in_filename;
     FILE *in_file;

     if (input_file != NULL) {
	  in_file = fopen(input_file, "r");
	  if (in_file == NULL) {
	       fprintf(stderr,
		       "cilk2c: unable to read input file \"%s\".\n", 
		       input_file);
	       exit(1);
	  }
	  in_filename = input_file;
	  SetFile(input_file, 0);
     } else {
	  if (stdin_name == NULL)
	       stdin_name = "stdin.cilk";
	  in_file = stdin;
	  in_filename = stdin_name;
	  SetFile(stdin_name, 0);
     }
     InputFileRoot = filename_root(in_filename);

     return (in_file);
}

/***********************************************************************\
 * Main
\***********************************************************************/

GLOBAL int main(int argc, char *argv[])
{
     Bool parsed_ok;

     Executable = argv[0];

     handle_options(argc, argv);

     InitTypes();
     init_symbol_tables(TRUE);
     InitOperatorTable();

     if (Parse) {
	  PhaseName = "Parsing";
	  yyin = get_input_file();
	  parsed_ok = yyparse();

	  if (Level != 0) {
	       SyntaxError("unexpected end of file");
	  }
     } else
	  Program = NULL;

     if (PrintSymTables) {
	  PrintSymbolTable(stdout, Externals);
     }
#ifndef NDEBUG
     if (Errors == 0) {
	  PhaseName = "Verification";
	  VerifyParse(Program);
     }
#endif

     PhaseName = "Semantic Check";
     if (Errors == 0 && SemanticCheck)
	  Program = SemanticCheckProgram(Program);

     PhaseName = "Analyze";
     if (Errors == 0 && Analyze)
	  AnalyzeProgram(Program);

     PhaseName = "Hint";
     if (Errors == 0 && DoHint)
	  HintProgram(Program);

     PhaseName = "Transform";
     if (Errors == 0 && Transform) {
	  Program = TransformProgram(Program);
     }
     PhaseName = "Output";
     if (Errors == 0 && GenerateOutput) {
	  if (output_file == NULL) {
	       OutputProgram(stdout, Program);
	  } else {
	       FILE *fp;

	       if ((fp = fopen(output_file, "w")) == NULL)
		    printf("Unable to open \"%s\" for output file\n", output_file);
	       else {
		    OutputProgram(fp, Program);
		    fclose(fp);
	       }
	  }
     }

     if (Errors > 0) {
	  fprintf(stderr, "\nCompilation Failed: %d error%s, %d warning%s\n",
		  Errors, PLURAL(Errors),
		  Warnings, PLURAL(Warnings));
     } else if (Warnings > 0) {
	  if (WarningLevel >= 5)
	       fprintf(stderr, "\nCompilation Successful (%d warning%s)\n",
		       Warnings,
		       PLURAL(Warnings));
     }
     if (PrintAST) {
	  PrintList(stdout, Program, -1);
	  fprintf(stdout, "\n");
     }

     return (Errors);
}
