/* $Id: cilk-internal.h,v 1.3 2001/01/22 23:11:34 matley Exp $ */
/*
 * Copyright (c) 2000 Massachusetts Institute of Technology
 * Copyright (c) 2000 Matteo Frigo
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/* common definitions used internally by the rts */
#if HAVE_CONFIG_H
#include <config.h>
#endif

#if HAVE_PTHREAD_H
#include <pthread.h>
#endif

/******************************************************
 *  Locks and associated operations
 ******************************************************/
typedef union {
     volatile int memory;
     pthread_mutex_t posix;
}  Cilk_mutex;

extern void Cilk_mutex_init(Cilk_mutex *lock);
extern void Cilk_mutex_wait(CilkWorkerState *const ws, Cilk_mutex *lock);
extern void Cilk_mutex_signal(Cilk_mutex *lock);
extern int  Cilk_mutex_try(Cilk_mutex *lock);

/******************************************************
 *  Closures and associated operations
 ******************************************************/
/*
 * If a frame is at the top of its processor stack, these additional
 * informations are recorded.
 */
enum ClosureStatus { CLOSURE_RUNNING = 42, CLOSURE_SUSPENDED,
                     CLOSURE_RETURNING, CLOSURE_READY };

/* The order is important here. */
enum AbortStatus { ABORT_ALL = 30 , ALMOST_NO_ABORT, NO_ABORT};

/*
 * This structure contains all the data needed in order to execute an
 * inlet.  Some inlets are "complete", other are not.  "Incomplete"
 * inlets are still waiting for data.
 *
 * Each closure has a list of complete inlets, and a list
 * of incomplete inlets.  The list of incomplete inlets
 * is isomorphic to the list of outstanding children,
 * and is also used for that purpose.
 * Complete inlets are ready to go, but are waiting for
 * atomic access to the frame.
 */
struct InletClosure {
     struct Closure *this;                /* child that structure is for */
     struct InletClosure *next;
  
     int index;
     void (*inlet) (CilkWorkerState *const, void *, void *, void *);
     int argsize;
     void *inlet_args;
};

/* 
 * the list of children is not distributed among
 * the children themselves, in order to avoid extra protocols
 * and locking.
 */
typedef struct Closure {
     Cilk_mutex mutex;          /* mutual exclusion lock */
     WHEN_CILK_DEBUG(int mutex_owner;)

     int join_counter; /* number of outstanding children */
     enum ClosureStatus status;
     
     /*
      * return value (and its size), valid only 
      * for CLOSURE_RETURNING closures 
      */
     void *return_value; 
     int return_size;

     struct InletClosure *complete_inlets;  /* runnable inlets */
     struct InletClosure *incomplete_inlets;  
     /* also, list of outstanding children */

     struct Closure *parent; /* the parent closure */

     CilkStackFrame *frame;       /* rest of the closure */
     enum AbortStatus abort_status; 
     CilkClosureCache *cache; /* see cilk.h */

     /* 
      * stuff related to ready deque.  These fields
      * must be managed only by the queue manager in sched.c
      */
     struct Closure *next_ready; 
     struct Closure *prev_ready; 
     WHEN_CILK_DEBUG(int owner_ready_deque;)

     /* miscellanea */
     WHEN_CILK_DEBUG(int magic;)
     /* indicates that this closure was allocated with mallec */
     WHEN_CILK_DEBUG(int malloced;)

     /* critical path and work */
     Cilk_time cp;      
     Cilk_time work;

     CILK_CACHE_LINE_PAD;
} Closure;

/* internal malloc */
extern void
 Cilk_internal_malloc_per_worker_init(CilkWorkerState *const ws);
extern void
 Cilk_internal_malloc_per_worker_cleanup(CilkWorkerState *const ws);
extern void Cilk_internal_malloc_global_init(void);
extern void Cilk_internal_malloc_global_cleanup(void);

/* barriers */
extern void Cilk_barrier(CilkWorkerState *const ws);
extern void Cilk_barrier_init(void);
extern void Cilk_barrier_per_worker_init(CilkWorkerState *const ws);

/* scheduler */
extern void Cilk_scheduler_init(void);
extern void Cilk_scheduler_per_worker_init(CilkWorkerState *const ws);
extern Closure *Cilk_create_initial_thread(
     int (*main)(CilkWorkerState *const ws, 
		 int argc, char *argv[]),
     int argc, char *argv[]);
extern void Cilk_scheduler(CilkWorkerState *const ws, Closure *t);
void Cilk_remove_and_free_closure_and_frame(
     CilkWorkerState *const ws, CilkStackFrame *f, int pn);
extern void Cilk_destroy_frame(
     CilkWorkerState *const ws, CilkStackFrame *f, size_t size);

/* exit protocol */
extern void Cilk_exit_from_user_main(
     CilkWorkerState *const ws, Closure *cl, int res);
extern void Cilk_really_exit(CilkWorkerState *const ws, int res);
extern void Cilk_really_exit_1(CilkWorkerState *const ws, int res);
extern void Cilk_die(const char *fmt,...);

/* memory allocator */
extern void *Cilk_malloc(size_t size);
extern void *Cilk_malloc_fixed(size_t size);
extern void *Cilk_memalign(size_t alignment, size_t size);
extern void Cilk_free(void *s);
extern void *Cilk_calloc(size_t nelem, size_t elsize);
extern void *Cilk_realloc(void *s, size_t size);
extern void *Cilk_valloc(size_t size);

/* other variables */
DEFINE_PARAMETER(extern int, yield_slice);
DEFINE_PARAMETER(extern int, StatLevel);
DEFINE_PARAMETER(extern int, trace_mode);
DEFINE_PARAMETER(extern int, dump_core);
DEFINE_PARAMETER(extern int, alloc_batch_size);
DEFINE_PARAMETER(extern size_t, stackdepth);
DEFINE_PARAMETER(extern int, memory_locks);

/* global scheduler state */
typedef struct {
     Cilk_time critical_path;
     Cilk_time total_work;
     int retval;
     int done;
     Cilk_mutex barrier_lock;
     volatile int barrier_counter;
     volatile int barrier_release;
     struct Cilk_im_descriptor 
          global_im_descriptor [CILK_INTERNAL_MALLOC_BUCKETS];
     struct Cilk_im_stats global_im_info;
     int im_allocated;
     int im_wasted;
     Cilk_mutex global_im_lock;
     char *global_pool_begin;
     char *global_pool_end;
     Cilk_mutex dprintf_lock;
     Cilk_mutex die_lock;
} CilkGlobalState;

/* 
 * Cilk_global_state contains all shared variables internal to the
 * runtime system
 */
extern CilkGlobalState *Cilk_global_state;

/* handy syntactic sugar */
#define USE_SHARED(id) (Cilk_global_state->id)


/* external malloc */
extern void Cilk_init_arena(size_t arenasize);

/* statistics */
#if CILK_STATS

enum { 
     EVENT_THREAD,

     EVENT_STEAL_ATTEMPT,
     EVENT_STEAL,
     EVENT_STEAL_NO_DEKKER,
     EVENT_STEAL_EMPTY_DEQUE,
     EVENT_STEAL_RETURNING,
     EVENT_STEAL_ABORT,

     EVENT_ABORT_SLOW,
     EVENT_ABORT_STANDALONE,
     EVENT_ABORT_READY_SUSPEND,
     EVENT_ABORT_READY_RETURN,

     EVENT_SUSPEND,
     EVENT_CILK_SYNC,

     EVENT_EXCEPTION,
     EVENT_EXCEPTION_STEAL,
     EVENT_EXCEPTION_ABORT,
     EVENT_EXCEPTION_ABORT_RETURN,
     EVENT_EXCEPTION_ABORT_SUSPEND,
     EVENT_EXCEPTION_OTHER,

     EVENT_RETURN_SLOW,
     EVENT_PROVABLY_GOOD_STEAL,
     EVENT_RETURN_ENQUEUE,
     EVENT_POLL_INLETS,
     EVENT_CLOSURE_CREATE,

     EVENT_IM_ALLOC_BATCH,
     EVENT_IM_FREE_BATCH,

     EVENT_USER0,
     EVENT_USER1,
     EVENT_USER2,
     EVENT_USER3,
     EVENT_USER4,
     EVENT_USER5,
     EVENT_USER6,
     EVENT_USER7,
     EVENT_NTYPES  		/* this must be last */
};
     
extern void Cilk_event(CilkWorkerState *const ws, int type);
extern void Cilk_event_gathering_init(void);
extern void Cilk_print_rts_statistics(void);
extern void Cilk_internal_malloc_print_statistics(void);

#else

#define Cilk_event(a, b)
#define Cilk_event_gathering_init()
#define Cilk_print_rts_statistics()
#define Cilk_internal_malloc_print_statistics()
#endif

/* timing */
#if CILK_TIMING
extern void Cilk_enter_state(CilkWorkerState *const ws, int state);
extern void Cilk_exit_state(CilkWorkerState *const ws, int state);
extern void Cilk_print_time_statistics(void);
extern void Cilk_print_detailed_time_statistics(void);
extern void Cilk_summarize_time_statistics(void);
extern double Cilk_compute_work(void);
extern void Cilk_time_gathering_init(void);

enum {
     STATE_TOTAL,
     STATE_WORKING,
     STATE_WAITING_FOR_LOCK,
     STATE_STEALING,
     STATE_RETURNING,
     STATE_ABORT_RECURSIVE,
     STATE_USER0,                /* for the user */
     STATE_USER1,
     STATE_USER2,
     STATE_USER3,
     STATE_USER4,
     STATE_USER5,
     STATE_USER6,
     STATE_USER7,
     TIME_NSTATES		/* this must be last */
};

#else

#define Cilk_enter_state(ws, state)
#define Cilk_exit_state(ws, state)
#define Cilk_print_time_statistics()
#define Cilk_print_detailed_time_statistics()
#define Cilk_summarize_time_statistics()
#define Cilk_compute_work() (0.0)
#define Cilk_time_gathering_init()

#endif				/* timing */

/*
 * Functions defined in the architecture-specific file and used
 * only by the runtime system.
 */
extern int Cilk_arch_specific_options(int i, int argc, char *argv[]);
extern void Cilk_arch_specific_init(void);
extern void Cilk_debug_init(void);
extern void Cilk_arch_specific_per_worker_init(void);
extern void Cilk_create_children(void (*child)(int));
extern void Cilk_lower_priority(CilkWorkerState *const ws);
extern void Cilk_raise_priority(CilkWorkerState *const ws);
extern void Cilk_yield(CilkWorkerState *const ws);
extern int Cilk_partition_size(void);

/*
 * functions used by the scheduler and by other protocols
 * operating on the Closures tree
 */
extern Closure *Cilk_Closure_create_malloc(CilkWorkerState *const ws);

/******************************************************
 *  Options processing
 ******************************************************/
typedef struct
{
     int nproc;
     int stackdepth;
     int statlevel;
     int yieldslice;
     char *infofile_name;  /* this should really be arch-specific */
     int dump_core;
     int alloc_batch_size;
     int memory_locks;
} Cilk_options;

/* command-line parser */
extern int Cilk_parse_command_line(
     Cilk_options *options, int *argc, char *argv[]);

#define CILK_DEFAULT_OPTIONS   (Cilk_options)	\
{						\
     1,						\
     CILK_DEFAULT_STACK_DEPTH,			\
     0,						\
     1,						\
     "-",					\
     0,						\
     64,					\
     0						\
}
