/*
 * $Id: cilk.c,v 1.2 2001/01/22 23:11:34 matley Exp $
 */
/*
 * Copyright (c) 2000 Massachusetts Institute of Technology
 * Copyright (c) 2000 Matteo Frigo
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <cilk.h>
#include <cilk-internal.h>
#include <stdio.h>

DEFINE_PARAMETER(static Cilk_time, start_time);

/* globally-visible options (read-only in child processes) */
DEFINE_PARAMETER(int, active_size);
DEFINE_PARAMETER(int, dump_core);
DEFINE_PARAMETER(int, StatLevel);
DEFINE_PARAMETER(int, yield_slice);
DEFINE_PARAMETER(size_t, stackdepth);
DEFINE_PARAMETER(char *, infofile_name);
			/* file where info messages are sent */

/*
 * HACK: this should be FILE *; but using a void * simplifies the
 * automatic generation of the file containing all shared/private/ro
 * variables.
 */
DEFINE_PARAMETER(void *, infofile);

/*
 * variables that children need to update (typically at the end
 * of the program
 */
DEFINE_PARAMETER(unsigned int, num_threads);
DEFINE_PARAMETER(unsigned int, num_steals);


/*
 * this string is printed when an assertion fails.  If we just inline
 * it, apparently gcc generates many copies of the string.
 */
DEFINE_PARAMETER(const char *, assertion_failed_msg);

DEFINE_PARAMETER(const char *, stack_overflow_msg);

static char *smart_sprint_time(double x)
{
     static char buf[128];

     if (x < 1.0E-3)
	  sprintf(buf, "%f us", x * 1.0E6);
     else if (x < 1.0)
	  sprintf(buf, "%f ms", x * 1.0E3);
     else
	  sprintf(buf, "%f s", x);

     return buf;
}

/*
 * The statistics printer 
 */
static void print_all_statistics(void)
{
     if (strcmp(USE_PARAMETER(infofile_name), "-") == 0)
	  USE_PARAMETER(infofile) = stderr;
     else
	  USE_PARAMETER(infofile) = 
	       fopen(USE_PARAMETER(infofile_name), "w");

     CILK_CHECK(USE_PARAMETER(infofile),
		(NULL,
		 "can't open %s for writing\n", 
		 USE_PARAMETER(infofile_name)));

     if (USE_PARAMETER(StatLevel) >= 1) {
	  /* Print Header line for statistics */
	  fprintf(USE_PARAMETER(infofile),
		  "\nRUNTIME SYSTEM STATISTICS:\n"
		  "\n");
	  /* level 1 and above: print wall clock, work and CP */
	  fprintf(USE_PARAMETER(infofile),
		  "Wall-clock running time on %d processor%s: %s\n",
		  USE_PARAMETER(active_size), 
		  USE_PARAMETER(active_size) > 1 ? "s" : "",
		  smart_sprint_time(Cilk_wall_time_to_sec(
		       Cilk_get_wall_time() - USE_PARAMETER(start_time))));
	  WHEN_CILK_TIMING({
	       fprintf(USE_PARAMETER(infofile), "Total work = %s\n",
		       smart_sprint_time(Cilk_compute_work()));
	  });

	  /*
	   * hack: we don't consider the total work if it is too
	   * small.  The trick allows us to compile the rts with
	   * CILK_CRITICAL_PATH=something and the user program with
	   * CILK_CRITICAL_PATH=0
	   */
	  WHEN_CILK_TIMING({
	       if (Cilk_time_to_sec(USE_SHARED(total_work)) > 0.001) {
		    fprintf(USE_PARAMETER(infofile),
			    "Total work (accumulated) = %s\n",
			    smart_sprint_time(
				 Cilk_time_to_sec(USE_SHARED(total_work))));
		    fprintf(USE_PARAMETER(infofile),
			    "Critical path = %s\n",
			    smart_sprint_time(
				 Cilk_time_to_sec(USE_SHARED(critical_path))));
		    fprintf(USE_PARAMETER(infofile),
			    "Parallelism = %f\n",
			    (double) USE_SHARED(total_work) /
			    (double) USE_SHARED(critical_path));
	       }
	  });
     }

     if (USE_PARAMETER(StatLevel) >= 2) {
	  /*
	   *  level 2 and above: print statistics on work stealing etc
	   */
	  Cilk_print_rts_statistics();
	  WHEN_CILK_TIMING({
	       if (Cilk_time_to_sec(USE_SHARED(total_work)) > 0.001
		   && USE_PARAMETER(num_threads) > 0) {
		    fprintf(USE_PARAMETER(infofile), 
			    "AVERAGE THREAD LENGTH = %s\n",
			    smart_sprint_time(
				 Cilk_time_to_sec(
				      USE_SHARED(total_work)) /
				 (double) USE_PARAMETER(num_threads)));
	       }

	       if (Cilk_time_to_sec(USE_SHARED(total_work)) > 0.001
		   && USE_PARAMETER(num_steals) > 0) {
		    fprintf(USE_PARAMETER(infofile),
			    "AVERAGE SUBCOMPUTATION LENGTH = %s\n",
			    smart_sprint_time(
				 Cilk_time_to_sec(
				      USE_SHARED(total_work)) /
				 ((double) USE_PARAMETER(num_steals) + 1.0)));
	       }
	  });
     }
     if (USE_PARAMETER(StatLevel) >= 3) {
	  /* level 3 and above: memory statistics */
	  Cilk_internal_malloc_print_statistics();
     }
     if (USE_PARAMETER(StatLevel) >= 4) {
	  /* level 4 and above: print work done by each processor */
	  Cilk_print_time_statistics();
     }
     if (USE_PARAMETER(StatLevel) >= 5) {
	  /* level 5 and above: print summary breakout of time */
	  Cilk_summarize_time_statistics();
     }
     if (USE_PARAMETER(StatLevel) >= 6) {
	  /* level 6 and above: print detailed breakout of time */
	  Cilk_print_detailed_time_statistics();
     }

     fclose(USE_PARAMETER(infofile));
}

/********************************************************
 * Runtime-system shared variables
 ********************************************************/
CilkGlobalState *Cilk_global_state;

/* 
 * allocate some chunk of shared memory for global variables
 */
void Cilk_create_global_state(void)
{
     Cilk_global_state = 
	  Cilk_malloc_fixed(sizeof (CilkGlobalState));
     CILK_CHECK(Cilk_global_state,
		(NULL, "Cannot allocate global state\n"));
}

static void init_variables(Cilk_options *options)
{
     /* initialize all shared_ro variables declared above */
     INIT_PARAMETER(start_time, Cilk_get_wall_time());
     INIT_PARAMETER(active_size, options->nproc);
     INIT_PARAMETER(dump_core, options->dump_core);
     INIT_PARAMETER(StatLevel, options->statlevel);
     INIT_PARAMETER(yield_slice, options->yieldslice);
     INIT_PARAMETER(stackdepth, options->stackdepth);
     INIT_PARAMETER(alloc_batch_size, options->alloc_batch_size);
     INIT_PARAMETER(infofile_name, options->infofile_name);
     INIT_PARAMETER(memory_locks, options->memory_locks);
     INIT_PARAMETER(infofile, (FILE *)0);
     INIT_PARAMETER(num_threads, 0);
     INIT_PARAMETER(num_steals, 0);
     INIT_PARAMETER(assertion_failed_msg,
		    "Assertion failed: %s line %d file %s\n"
		    "This is either a Cilk bug, or your program\n"
		    "has corrupted some Cilk internal data structure.\n");
     INIT_PARAMETER(stack_overflow_msg,
		    "Cilk runtime system: Stack overflow.\n"
		    "Your program may have infinite recursion.\n");
}

/* global init */
static void Cilk_global_init(Cilk_options *options)
{
     init_variables(options);

     /* setup the address space */
     Cilk_create_global_state();

     USE_SHARED(critical_path) = 0;
     USE_SHARED(total_work) = 0;

     /* debug module */
     Cilk_debug_init();

     /* architecture-specific module */
     Cilk_arch_specific_init();

     /* initialize barrier code */
     Cilk_barrier_init();

     /* initialize the statistics gatherer and timer */
     Cilk_event_gathering_init();
     Cilk_time_gathering_init();

     /* run initialization hooks for library functions */
     Cilk_run_hooks(Cilk_init_global_hooks);
}

DEFINE_PARAMETER(static struct Closure *, invoke_main);

/*
 * this function is the entry point for child processes
 */
static void Cilk_child_main(int id)
{
     CilkWorkerState *const ws = Cilk_malloc_fixed(sizeof(CilkWorkerState));

     ws->self = id;
     Cilk_barrier_per_worker_init(ws);

     Cilk_arch_specific_per_worker_init();

     /* initialize scheduler per worker */
     Cilk_scheduler_per_worker_init(ws);

     /* run initialization hooks for library functions */
     Cilk_run_hooks(Cilk_init_per_worker_hooks);

     Cilk_barrier(ws);
     
     if (id == 0)
	  Cilk_scheduler(ws, USE_PARAMETER(invoke_main));
     else
	  Cilk_scheduler(ws, (Closure *) NULL);
}

int Cilk_start(int (*main)(CilkWorkerState *const ws,
			   int argc, char *argv[]), 
	       int argc, char *argv[])
{
     Cilk_options options;

     /* parse command-line arguments */
     if (Cilk_parse_command_line(&options, &argc, argv))
	  return 0;

     /* initialize global part of system */
     Cilk_global_init(&options);

     /* initialize the scheduler */
     Cilk_scheduler_init();
     USE_PARAMETER(invoke_main) =
	  Cilk_create_initial_thread(main, argc, argv);

     /* create children; when this returns we are done */
     Cilk_create_children(Cilk_child_main);

     Cilk_internal_malloc_global_cleanup();

     /* everybody is done: print stats and exit */
     print_all_statistics();
     exit(USE_SHARED(retval));
     return -1;
}

void Cilk_really_exit_1(CilkWorkerState *const ws, int res) {
     Cilk_exit_from_user_main(
	  ws, USE_PARAMETER(invoke_main), res);
}

/* magic consistency check */
int CILK_MAGIC_NAME;
